#!/usr/bin/python3
import yaml, sqlite3
import os, re, shutil
import logging, traceback

import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib

logger = None

# 创建一个DBus服务类
class KYSettingsDBusService(dbus.service.Object):
    __read_only_view = {}
    __root_view = {}

    def __init__(self, bus_name, object_path):
        logger.info(f'DBus Class init')
        dbus.service.Object.__init__(self, bus_name, object_path)

        os.makedirs('/etc/kylin-config', exist_ok=True)

        self.__reload__()

        logger.info(f'DBus Class init end')

    # 重载
    def __reload__(self):
        self.__update_read_only_db__()
        self.__update_root_db__()
        return 1

    # 更新初始数据库
    def __update_read_only_db__(self):
        path = '/etc/kylin-config/user.db'
        is_new_db = os.path.exists(path)

        try:
            if not is_new_db:
                conn = sqlite3.connect(path)
                cursor = conn.cursor()

                conn.execute('BEGIN TRANSACTION')
                cursor.execute('''CREATE TABLE IF NOT EXISTS app
                            (id INTEGER PRIMARY KEY AUTOINCREMENT,
                            app_name TEXT,
                            default_version TEXT)''')
                cursor.execute('''CREATE TABLE IF NOT EXISTS version
                               (id INTEGER PRIMARY KEY AUTOINCREMENT,
                               app_id INTEGER,
                               version TEXT,
                               compatible TEXT,
                               FOREIGN KEY (app_id) REFERENCES app(id))''')
                cursor.execute('''CREATE TABLE IF NOT EXISTS configures
                               (id INTEGER PRIMARY KEY AUTOINCREMENT,
                               version_id INTEGER,
                               node_name TEXT,
                               node_type TEXT,
                               permission TEXT,
                               description TEXT,
                               summary TEXT,
                               parent INTEGER,
                               value_type TEXT,
                               custom_value TEXT,
                               default_value TEXT,
                               range TEXT,
                               extends TEXT,
                               FOREIGN KEY (version_id) REFERENCES version(id),
                               FOREIGN KEY (parent) REFERENCES configures(id))''')
                conn.commit()
                conn.close()
            self.__load_read_only_view__()
            self.__dict2db__(self.__read_only_view, path)
            return True
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__update_read_only_db__ {err_msg}')
            return False
    
    # 读取配置文件
    def __load_read_only_view__(self):
        logger.info('Read conifg file')

        self.__read_only_view.clear()

        with open('/etc/kylin-config/conf2.yaml', 'r') as file:
            configures = yaml.safe_load(file)

        for dir in configures['dirs']:
            list = self.__sort_diretory__(f'/etc/kylin-config/{dir}')
            for file in list:
                # logger.info('Read conifg file')
                data = self.__load_one_file__(f'/etc/kylin-config/{dir}/{file}')
                # logger.info('Read conifg file')

                if not data:
                    continue
                
                # logger.info('Merger in view')
                is_group_file = False
                for value in data.values():
                    if isinstance(value, dict):
                        is_group_file = True if '_default' in next(iter(value.values())) else False
                        break
                if is_group_file:
                    self.__common_override__(self.__read_only_view, data)
                else:
                    self.__merge_dict__(self.__read_only_view, data)
                # logger.info('Merger in view end')
        logger.info('Read conifg file end')

    # 按规范对文件名进行排序
    def __sort_diretory__(self, yaml_dir):
        if not os.path.exists(yaml_dir):
            return []

        # 获取目录内所有文件名
        file_list = os.listdir(yaml_dir)

        # 定义排序函数
        def custom_sort(file_name):
            match = re.match(r"(\d+)?-?(.*)", file_name)
            if match.group(1):
                return (int(match.group(1)), match.group(2))
            else:
                return (0, match.group(2))

        sorted_file_list = sorted(file_list, key=custom_sort)
        return sorted_file_list
    
    # 读取一个文件
    def __load_one_file__(self, file:str):
        chinese_pattern = re.compile(u'[\u4e00-\u9fa5]+')
        if bool(chinese_pattern.search(file)) or file.startswith('.'):
            return {}

        yaml_data = {}
        if file.endswith('.yaml'):
            with open(file,'r') as stream:
                yaml_data = yaml.safe_load(stream)
            if yaml_data is None:
                logger.error(f'Load {file} failed. Check the text formatting')
                return {}
        return yaml_data
    
    # 合并两个字典
    def __merge_dict__(self, dest:dict , src:dict):
        if src == {}:
            return
        if dest == {}:
            dest.update(src)
            return
        for key, value in src.items():
            if key in dest and isinstance(dest[key], dict) and isinstance(src[key], dict):
                self.__merge_dict__(dest[key], src[key])
            else:
                dest[key] = value
        return

    # 全局配置字典覆盖
    def __common_override__(self, dest:dict, src:dict):
        for app in dest:
            for version in dest[app]:
                for group in dest[app][version]:
                    if group in src:
                        self.__merge_dict__(dest[app][version][group], src[group])
    
    # 字典写入到数据库
    def __dict2db__(self, dict:dict, path:str):
        try:
            conn = sqlite3.connect(path)
            cursor = conn.cursor()

            conn.execute('BEGIN TRANSACTION')
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
            tables = cursor.fetchall()

            # 遍历所有表名，对每个表执行DELETE语句以删除所有行
            for table in tables:
                cursor.execute(f"DELETE FROM {table[0]}")

            version_id = 0
            apps = list(dict.keys())
            for app_id in range(0, len(apps)):
                app = apps[app_id]
                default_version = self.__calculate_default_version__(dict[app])
                cursor.execute("INSERT INTO app (app_name, default_version) VALUES (?,?)",(app, default_version))

                versions = list(dict[app].keys())
                for i in range(0, len(versions)):
                    version = versions[i]
                    if version == '_default_version':
                        continue
                    version_id = version_id + 1
                    cursor.execute("INSERT INTO version (app_id, version, compatible) VALUES (?, ?, ?)",
                                (app_id + 1, version, dict[app][version].get('compatible')))
                    
                    configures = dict[app][version]
                    self.__write_group_in_db__(configures, configures, cursor, version_id, 0)
            conn.commit()
            conn.close()
            logger.info('write config in database end')
            return True
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__dict2db__ {err_msg}')
            conn.rollback()
            conn.commit()
            return False

    # 计算默认版本
    def __calculate_default_version__(self, data:dict):
        versions = list(data.keys())
        if '_default_version' in versions:
            return data.get('_default_version')
        default_version = ''
        for version in versions:
           if default_version == '':
               default_version = version
               continue
           default_nums = re.findall(r'\d+', default_version)
           nums = re.findall(r'\d+', version)
           for i in range(0, 6):
               if nums[i] > default_nums[i]:
                   default_version = version
                   break
        return default_version
    
    #递归写入组到configures表
    def __write_group_in_db__(self, version_data:dict, data: dict, cursor: sqlite3.Cursor, version_id, parent_id):
        groups = []
        for key, value in data.items():
            if isinstance(value, dict):
                if(value.get('_default') is None):
                    groups.append(key)
                else:
                    self.__write_key_in_db__(version_data, key, value, cursor, version_id, parent_id)

        for group in groups:
            node_name = group

            node_type = 'schema'

            permission = data[group].get('_permission')

            description = data[group].get('_description')

            summary = data[group].get('_summary')
            
            extends = data[group].get('_extends')

            cursor.execute('''INSERT INTO configures 
                           (version_id, node_name, node_type, permission, description, summary, parent, extends) 
                           VALUES (?, ?, ?, ?, ?, ?, ?, ?)''',
                           (version_id, node_name, node_type, permission, description, summary, parent_id, extends))
            group_id = cursor.lastrowid
            self.__write_group_in_db__(version_data, data[group], cursor, version_id, group_id)

    # 写入键的数据到configures表内
    def __write_key_in_db__(self, version_data:dict, key:str, value:dict, cursor: sqlite3.Cursor, version_id, parent_id):
        node_name = key

        node_type = 'key'

        permission = value.get('_permission')
        if permission is None:
            permission = 'public'

        description = value.get('_description')

        summary = value.get('_summary')

        value_type = value.get('_type')

        custom_value = value.get('_value')

        default_value = value.get('_default')
        if isinstance(default_value, bool):
            default_value = 'true' if default_value else 'false'
        
        range = value.get('_range')
        if value_type == 'enum' and isinstance(range, str) and range.startswith('@'):
            enum_range = {}
            element_list = self.__recursive_search__(version_data, value['_range'][1:])
            for element in element_list:
                num = element['_value']
                if isinstance(num, str):
                    if num.startswith('0x'):
                        num = int(num, 16)
                    else:
                        num = int(num, 10)
                enum_range[element['_nick']] = num
            range = str(enum_range)
        print(range)

        cursor.execute('''INSERT INTO configures 
                       (version_id, node_name, node_type, permission, description, summary, parent, value_type, custom_value, default_value, range) 
                       VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)''',
                       (version_id, node_name, node_type, permission, description, summary, parent_id, value_type, custom_value, default_value, range))

    # 查询枚举类型数据的枚举定义
    def __recursive_search__(self, dictionary, key):
        if key in dictionary:
            return dictionary[key]
        else:
            for value in dictionary.values():
                if isinstance(value, dict):
                    result = self.__recursive_search__(value, key)
                    if result is not None:
                        return result
        return None
        
    # 更新root用户数据库
    def __update_root_db__(self):
        try:
            path = '/root/.config/kylin-config/user.db'
            ro_path = '/etc/kylin-config/user.db'
            if not os.path.exists(path):
                os.makedirs(os.path.dirname(path), exist_ok=True)
                shutil.copyfile(ro_path, path)
                self.__root_view = self.__db2dict__(path)
            else:
                self.__root_view = self.__db2dict__(path)
                self.__del_discarded_key__(self.__root_view, self.__read_only_view)
                self.__merge_dict__(self.__root_view, self.__read_only_view)
                self.__dict2db__(self.__root_view, path)
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__update_root_db__ {err_msg}')

    # 读数据库数据到字典
    def __db2dict__(self, path):
        try:
            conn = sqlite3.connect(path)
            cursor = conn.cursor()

            cursor.execute('SELECT * FROM app')
            apps = cursor.fetchall()

            cursor.execute('SELECT * FROM version')
            versions = cursor.fetchall()

            cursor.execute('SELECT * FROM configures')
            configures = cursor.fetchall()
            conn.close()

            data = {}
            stack = []
            i = j = k = 0
            for i in range(len(apps)):
                data[apps[i][1]] = {}
                if(apps[i][2]):
                    data[apps[i][1]]['_default_version'] = apps[i][2]
                for j in range(j, len(versions)):
                    # app_id改变，退出版本循环
                    if versions[j][1] != i + 1:
                        break

                    data[apps[i][1]][versions[j][2]] = {}
                    if versions[j][3]:
                        data[apps[i][1]][versions[j][2]]['compatible'] = versions[j][3]

                    stack.append((0, data[apps[i][1]][versions[j][2]]))
                    # 如果占用内存过大，可以改为在这里读取对应版本的所有配置
                    # 不再保存k值，每次循环从0开始
                    for k in range(k, len(configures)):
                        # version_id改变，退出循环
                        if configures[k][1] != j + 1:
                            stack.clear()
                            break

                        if configures[k][7] != stack[-1][0]:
                            stack.pop()
                        
                        node_name = configures[k][2]
                        node_type = configures[k][3]
                        permission = configures[k][4]
                        description = configures[k][5]
                        summary = configures[k][6]
                        value_type = configures[k][8]
                        custom_value = configures[k][9]
                        default_value = configures[k][10]
                        value_range = configures[k][11]
                        extends = configures[k][12]

                        if node_type == 'schema':
                            tmp = stack[-1][1]
                            tmp[node_name] = {}
                            stack.append((configures[k][0], tmp[node_name]))
                            if permission is not None:
                                tmp[node_name]['_permission'] = permission
                            if description is not None:
                                tmp[node_name]['_description'] = description
                            if summary is not None:
                                tmp[node_name]['_summary'] = summary
                            if extends is not None:
                                tmp[node_name]['_extends'] = extends
                        elif node_type == 'key':
                            tmp = stack[-1][1]
                            tmp[node_name] = {}
                            if permission is not None:
                                tmp[node_name]['_permission'] = permission
                            if description is not None:
                                tmp[node_name]['_description'] = description
                            if summary is not None:
                                tmp[node_name]['_summary'] = summary
                            if value_type is not None:
                                tmp[node_name]['_type'] = value_type
                            if custom_value is not None:
                                tmp[node_name]['_value'] = custom_value
                            if default_value is not None:
                                tmp[node_name]['_default'] = default_value
                            if value_range is not None:
                                tmp[node_name]['_range'] = value_range
                            
                            
            return data
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__db2dict__ {err_msg}')
            return {}
    
    # 删除dest中存在，但不存在与src中的键或组
    def __del_discarded_key__(self, dest:dict, src:dict):
        keys_to_delete = []
        for key in dest:
            if not isinstance(dest[key], dict):
                continue
            if key not in src:
                if dest[key].get('_extends') is None:
                    keys_to_delete.append(key)
            else:
                self.__del_discarded_key__(dest[key], src[key])

        for key in keys_to_delete:
            dest.pop(key)

    # 导出时，使用value的值替带default值
    def __value_override_default__(self, dest:dict, src:dict):
        for key, value in src.items():
            if key not in dest:
                continue
            if isinstance(src[key], dict):
                self.__value_override_default__(dest[key], src[key])
            else:
                if src.get('_value') is not None:
                    dest['_default'] = src['_value']
                return

    # key值改变信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='ssss')
    def key_changed(self, id, version, key, user):
        pass

    # 配置信息update信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='')
    def updated(self):
        pass

    # 设置一个配置的值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='sssss', out_signature='i')
    def set(self, user, id, version, key, value):
        try:
            logger.info(f'set begin')
            # 修改用户配置文件，写入到数据库
            group_list = id.split('.')
            sql_path = ''
            if user == 'root':
                sql_path = '/root/.config/kylin-config/user.db'
            else:
                sql_path = f'/home/{user}/.config/kylin-config/user.db'

            conn = sqlite3.connect(sql_path)
            cursor = conn.cursor()

            # 开始写入事务
            conn.execute('BEGIN TRANSACTION')

            cursor.execute("SELECT id FROM app WHERE app_name = ?", (group_list[0],))
            app_id = cursor.fetchone()[0]

            cursor.execute("SELECT id FROM version WHERE app_id = ? AND version = ?", (app_id, version))
            version_id = cursor.fetchone()[0]

            # 查找组
            parent_id = 0
            for group_name in group_list[1:]:
                cursor.execute("SELECT id FROM configures WHERE version_id = ? AND parent = ? AND node_name = ? AND node_type = 'schema'", (version_id, parent_id, group_name))
                parent_id = cursor.fetchone()[0]

            cursor.execute("UPDATE configures SET custom_value = ? WHERE version_id = ? AND parent = ? AND node_name = ? AND node_type = 'key'", (value, version_id, parent_id, key))

            conn.commit()
            conn.close()

            self.key_changed(id, version, key, user)

            if user == 'root':
                self.__root_view = self.__db2dict__('/root/.config/kylin-config/user.db')

            logger.info(f'set begin')

            return 1
        except Exception as e:
            conn.rollback()
            conn.close()
            err_msg = traceback.format_exc()
            logger.error(f'set {err_msg}')
            return 0
    
    # 恢复单个默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ssss', out_signature='i')
    def reset(self,user, id, version, key):
        try:
            logger.info(f'reset begin')
            # 修改用户配置文件，写入到数据库
            group_list = id.split('.')
            sql_path = ''
            if user == 'root':
                sql_path = '/root/.config/kylin-config/user.db'
            else:
                sql_path = f'/home/{user}/.config/kylin-config/user.db'

            conn = sqlite3.connect(sql_path)
            cursor = conn.cursor()

            # 开始写入事务
            conn.execute('BEGIN TRANSACTION')

            cursor.execute("SELECT id FROM app WHERE app_name = ?", (group_list[0],))
            app_id = cursor.fetchone()[0]

            cursor.execute("SELECT id FROM version WHERE app_id = ? AND version = ?", (app_id,version))
            version_id = cursor.fetchone()[0]

            # 查找组
            parent_id = 0
            for group_name in group_list[1:]:
                cursor.execute("SELECT id FROM configures WHERE version_id = ? AND parent = ? AND node_name = ? AND node_type = 'schema'", (version_id, parent_id, group_name))
                parent_id = cursor.fetchone()[0]

            cursor.execute("UPDATE configures SET custom_value = NULL WHERE version_id = ? AND parent = ? AND node_name = ? AND node_type = 'key'", (version_id, parent_id, key))

            conn.commit()
            conn.close()

            self.key_changed(id,version,key, user)

            if user == 'root':
                self.__root_view = self.__db2dict__('/root/.config/kylin-config/user.db')
            
            logger.info(f'reset end')
            
            return 1
        except Exception as e:
            conn.rollback()
            conn.close()
            err_msg = traceback.format_exc()
            logger.error(f'reset {err_msg}')
            return 0

    # 恢复整个id的默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='', out_signature='i')
    def reload(self):
        result = self.__reload__()
        if result:
            self.updated()
        return result

    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ss', out_signature='b')
    def save(self, user, path):
        try:
            logger.info(f'save begin')

            if not os.path.exists(path):
                logger.error(f'{path} not exists')
                return False
            
            data = {}
            if 'root' == user:
                data = self.__db2dict__('/root/.config/kylin-config/user.db')
            else:
                data = self.__db2dict__(f'/home/{user}/.config/kylin-config/user.db')

            if data == {}:
                logger.error(f'read sqlite failed')
                return False
            
            tmp = {}
            tmp.update(self.__read_only_view)
            if tmp == {} or tmp is None:
                logger.error(f'read only view is None')
                return False
            #将用户配置写入到复制的只读视图里
            self.__value_override_default__(tmp, data)

            for key in tmp:
                data = {}
                data[key] = tmp[key]
                with open(f'{path}/{key}.yaml', 'w') as yaml_file:
                    yaml_file.write(yaml.safe_dump(data,allow_unicode = True))

            logger.info(f'save end')
            return True
        except Exception as e:
            err_msg = traceback.format_exc()
            logger.error(f'save {err_msg}')

    #复制old_id的配置到new_id的位置，并将新的数据写入到数据库
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ssss', out_signature='b')
    def extends_id(self, old_id:str, new_id:str, version:str, user:str):
        try:
            logger.info(f'extends_id begin')
            if user == 'root':
                target_view = self.__root_view
                target_file = "/root/.config/kylin-config/user.db"
            else:
                target_file = f'/home/{user}/.config/kylin-config/user.db'
                target_view = self.__db2dict__(target_file)
            
            #获取app的引用
            list = old_id.split('.', -1)
            app = target_view.get(list[0])

            if app is None:
                logger.error(f'{old_id} is not exists')
                return False
            
            #获取目标版本的引用，如果为空，获取默认版本
            if version == '':
                old_data = app[app['_default_version']]
            else:
                old_data = app.get(version)

            if old_data is None:
                logger.error(f'{version} is not invalid')

            for iter in list[1:]:
                old_data = old_data.get(iter)
                if old_data is None:
                    logger.error(f'{old_id} is not exists')
                    return False
                
            tmp = {}
            tmp.update(old_data)

            list = new_id.split('.', -1)
            new_data = target_view.get(list[0])
            if new_data != app:
                logger.error(f'{new_id} is invalid')

            if version == '':
                new_data = app[app['_default_version']]
            else:
                new_data = app.get(version)
            
            for iter in list[1:]:
                if iter not in new_data:
                    new_data[iter] = {}
                new_data = new_data[iter]
            
            new_data.update(tmp)
            new_data['_extends'] = old_id
            logger.info(f'extends_id end')
            return self.__dict2db__(target_view, target_file)
        
        except Exception as e:
            err_msg = traceback.format_exc()
            logger.error(f'extends_id {err_msg}')

if __name__ == "__main__":
    logging.basicConfig(filename='/etc/kylin-config/systembus.log', 
                        level=logging.DEBUG, 
                        format='%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s', 
                        datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()

    # 初始化DBus主循环
    DBusGMainLoop(set_as_default=True)

    # 连接到DBus系统总线
    bus = dbus.SystemBus()

    # 注册一个DBus服务
    bus_name = dbus.service.BusName("com.kylin.kysdk.conf2", bus=bus)
    object_path = "/com/kylin/kysdk/conf2"
    my_dbus_service = KYSettingsDBusService(bus_name, object_path)

    # 开始DBus主循环
    mainloop = GLib.MainLoop()
    mainloop.run()