#!/usr/bin/python3
import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib

import os, shutil, re
import sqlite3
import logging, traceback

logger = None
destination_file = '~/.config/kylin-config/user.db'

class KYSettingsDBusSessionService(dbus.service.Object):
    __user_view = {}

    def __init__(self, bus_name, object_path):
        logger.info('Session DBus Init')
        dbus.service.Object.__init__(self, bus_name, object_path)

        conn = dbus.SystemBus()
        conn.add_signal_receiver(self.__reload_handler__, 
                            signal_name='updated',
                            dbus_interface='com.kylin.kysdk.conf2',
                            bus_name='com.kylin.kysdk.conf2',
                            path="/com/kylin/kysdk/conf2")
        
        conn.add_signal_receiver(self.__key_changed_handler, 
                    signal_name='key_changed',
                    dbus_interface='com.kylin.kysdk.conf2',
                    bus_name='com.kylin.kysdk.conf2',
                    path="/com/kylin/kysdk/conf2")
        self.Init()
        logger.info('Session DBus end')

    def __reload_handler__(self):
        logger.info('On update')
        self.Init()
        self.updated()

    def __key_changed_handler(self, id, version, key, user):
        logger.info('On key changed')
        self.__user_view = self.__db2dict__(destination_file)
        self.key_changed(id, version, key, user)

    # 更新root用户数据库
    def __update_user_db__(self):
        try:
            ro_path = '/etc/kylin-config/user.db'
            if not os.path.exists(destination_file):
                os.makedirs(os.path.dirname(destination_file), exist_ok=True)
                shutil.copyfile(ro_path, destination_file)
                self.__user_view = self.__db2dict__(destination_file)
            else:
                data = self.__db2dict__(ro_path)
                self.__del_discarded_key__(self.__user_view, data)
                self.__merge_dict__(self.__user_view, data)
                self.__dict2db__(self.__user_view, destination_file)
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__update_root_db__ {err_msg}')

    # 合并两个字典
    def __merge_dict__(self, dest:dict , src:dict):
        if src == {}:
            return
        if dest == {}:
            dest.update(src)
            return
        for key, value in src.items():
            if key in dest and isinstance(dest[key], dict) and isinstance(src[key], dict):
                self.__merge_dict__(dest[key], src[key])
            else:
                dest[key] = value
        return
    
    # 字典写入到数据库
    def __dict2db__(self, dict:dict, path:str):
        try:
            conn = sqlite3.connect(path)
            cursor = conn.cursor()

            conn.execute('BEGIN TRANSACTION')
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
            tables = cursor.fetchall()

            # 遍历所有表名，对每个表执行DELETE语句以删除所有行
            for table in tables:
                cursor.execute(f"DELETE FROM {table[0]}")

            version_id = 0
            apps = list(dict.keys())
            for app_id in range(0, len(apps)):
                app = apps[app_id]
                default_version = self.__calculate_default_version__(dict[app])
                cursor.execute("INSERT INTO app (app_name, default_version) VALUES (?,?)",(app, default_version))

                versions = list(dict[app].keys())
                for i in range(0, len(versions)):
                    version = versions[i]
                    if version == '_default_version':
                        continue
                    version_id = version_id + 1
                    cursor.execute("INSERT INTO version (app_id, version, compatible) VALUES (?, ?, ?)",
                                (app_id + 1, version, dict[app][version].get('compatible')))
                    
                    configures = dict[app][version]
                    self.__write_group_in_db__(configures, configures, cursor, version_id, 0)
            conn.commit()
            conn.close()
            return True
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__dict2db__ {err_msg}')
            conn.rollback()
            conn.commit()
            return False
        
        # 计算默认版本
    def __calculate_default_version__(self, data:dict):
        versions = list(data.keys())
        if '_default_version' in versions:
            return data.get('_default_version')
        default_version = ''
        for version in versions:
           if default_version == '':
               default_version = version
               continue
           default_nums = re.findall(r'\d+', default_version)
           nums = re.findall(r'\d+', version)
           for i in range(0, 6):
               if nums[i] > default_nums[i]:
                   default_version = version
                   break
        return default_version
    
    #递归写入组到configures表
    def __write_group_in_db__(self, version_data:dict, data: dict, cursor: sqlite3.Cursor, version_id, parent_id):
        groups = []
        for key, value in data.items():
            if isinstance(value, dict):
                if(value.get('_default') is None):
                    groups.append(key)
                else:
                    self.__write_key_in_db__(version_data, key, value, cursor, version_id, parent_id)

        for group in groups:
            node_name = group

            node_type = 'schema'

            permission = data[group].get('_permission')

            description = data[group].get('_description')

            summary = data[group].get('_summary')
            
            extends = data[group].get('_extends')

            cursor.execute('''INSERT INTO configures 
                           (version_id, node_name, node_type, permission, description, summary, parent, extends) 
                           VALUES (?, ?, ?, ?, ?, ?, ?, ?)''',
                           (version_id, node_name, node_type, permission, description, summary, parent_id, extends))
            group_id = cursor.lastrowid
            self.__write_group_in_db__(version_data, data[group], cursor, version_id, group_id)

    # 写入键的数据到configures表内
    def __write_key_in_db__(self, version_data:dict, key:str, value:dict, cursor: sqlite3.Cursor, version_id, parent_id):
        node_name = key

        node_type = 'key'

        permission = value.get('_permission')
        if permission is None:
            permission = 'public'

        description = value.get('_description')

        summary = value.get('_summary')

        value_type = value.get('_type')

        custom_value = value.get('_value')

        default_value = value.get('_default')
        if isinstance(default_value, bool):
            default_value = 'true' if default_value else 'false'
        
        range = value.get('_range')
        if value_type == 'enum' and isinstance(range, str) and range.startswith('@'):
            enum_range = {}
            element_list = self.__recursive_search__(version_data, value['_range'][1:])
            for element in element_list:
                num = element['_value']
                if isinstance(num, str):
                    if num.startswith('0x'):
                        num = int(num, 16)
                    else:
                        num = int(num, 10)
                enum_range[element['_nick']] = num
            range = str(enum_range)

        cursor.execute('''INSERT INTO configures 
                       (version_id, node_name, node_type, permission, description, summary, parent, value_type, custom_value, default_value, range) 
                       VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)''',
                       (version_id, node_name, node_type, permission, description, summary, parent_id, value_type, custom_value, default_value, range))

    # 查询枚举类型数据的枚举定义
    def __recursive_search__(self, dictionary, key):
        if key in dictionary:
            return dictionary[key]
        else:
            for value in dictionary.values():
                if isinstance(value, dict):
                    result = self.__recursive_search__(value, key)
                    if result is not None:
                        return result
        return None

    # 读数据库数据到字典
    def __db2dict__(self, path):
        try:
            conn = sqlite3.connect(path)
            cursor = conn.cursor()

            cursor.execute('SELECT * FROM app')
            apps = cursor.fetchall()

            cursor.execute('SELECT * FROM version')
            versions = cursor.fetchall()

            cursor.execute('SELECT * FROM configures')
            configures = cursor.fetchall()
            conn.close()

            data = {}
            stack = []
            i = j = k = 0
            for i in range(len(apps)):
                data[apps[i][1]] = {}
                if(apps[i][2]):
                    data[apps[i][1]]['_default_version'] = apps[i][2]
                for j in range(j, len(versions)):
                    # app_id改变，退出版本循环
                    if versions[j][1] != i + 1:
                        break

                    data[apps[i][1]][versions[j][2]] = {}
                    if versions[j][3]:
                        data[apps[i][1]][versions[j][2]]['compatible'] = versions[j][3]

                    stack.append((0, data[apps[i][1]][versions[j][2]]))
                    # 如果占用内存过大，可以改为在这里读取对应版本的所有配置
                    # 不再保存k值，每次循环从0开始
                    for k in range(k, len(configures)):
                        # version_id改变，退出循环
                        if configures[k][1] != j + 1:
                            stack.clear()
                            break

                        if configures[k][7] != stack[-1][0]:
                            stack.pop()
                        
                        node_name = configures[k][2]
                        node_type = configures[k][3]
                        permission = configures[k][4]
                        description = configures[k][5]
                        summary = configures[k][6]
                        value_type = configures[k][8]
                        custom_value = configures[k][9]
                        default_value = configures[k][10]
                        value_range = configures[k][11]
                        extends = configures[k][12]

                        if node_type == 'schema':
                            tmp = stack[-1][1]
                            tmp[node_name] = {}
                            stack.append((configures[k][0], tmp[node_name]))
                            if permission is not None:
                                tmp[node_name]['_permission'] = permission
                            if description is not None:
                                tmp[node_name]['_description'] = description
                            if summary is not None:
                                tmp[node_name]['_summary'] = summary
                            if extends is not None:
                                tmp[node_name]['_extends'] = extends
                        elif node_type == 'key':
                            tmp = stack[-1][1]
                            tmp[node_name] = {}
                            if permission is not None:
                                tmp[node_name]['_permission'] = permission
                            if description is not None:
                                tmp[node_name]['_description'] = description
                            if summary is not None:
                                tmp[node_name]['_summary'] = summary
                            if value_type is not None:
                                tmp[node_name]['_type'] = value_type
                            if custom_value is not None:
                                tmp[node_name]['_value'] = custom_value
                            if default_value is not None:
                                tmp[node_name]['_default'] = default_value
                            if value_range is not None:
                                tmp[node_name]['_range'] = value_range
            return data
        except:
            err_msg = traceback.format_exc()
            logger.error(f'__db2dict__ {err_msg}')
            return {}
    
    # 删除dest中存在，但不存在与src中的键或组
    def __del_discarded_key__(self, dest:dict, src:dict):
        keys_to_delete = []
        for key in dest:
            if not isinstance(dest[key], dict):
                continue
            if key not in src:
                if dest[key].get('extends') is None:
                    keys_to_delete.append(key)
            else:
                self.__del_discarded_key__(dest[key], src[key])

        for key in keys_to_delete:
            dest.pop(key)

    # key值改变信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='ssss')
    def key_changed(self, id, version, key, user):
        logger.info(f'emit {key} changed')

    # 重载信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='')
    def updated(self):
        logger.info('emit updated')

    @dbus.service.method("com.kylin.kysdk.conf2")
    def Init(self):
        self.__update_user_db__()

if __name__ == "__main__":
    destination_file = os.path.expanduser(destination_file)
    logging.basicConfig(filename=destination_file.replace('user.db', 'sessionbus.log'), 
                        level=logging.DEBUG, 
                        format='%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s', 
                        datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()
    
    DBusGMainLoop(set_as_default=True)

    bus_session = dbus.SessionBus()
    object_path = '/com/kylin/kysdk/conf2'
    bus_sesssion_name = dbus.service.BusName("com.kylin.kysdk.conf2", bus = bus_session)
    my_dbus_session_service = KYSettingsDBusSessionService(bus_sesssion_name, object_path)

    mainloop = GLib.MainLoop()
    mainloop.run()